const React = require('react');
const { connect } = require('react-redux');
const Setting = require('lib/models/Setting');
const EncryptionService = require('lib/services/EncryptionService');
const { Header } = require('./Header.min.js');
const { themeStyle } = require('../theme.js');
const { _ } = require('lib/locale.js');
const { time } = require('lib/time-utils.js');
const dialogs = require('./dialogs');
const shared = require('lib/components/shared/encryption-config-shared.js');
const { bridge } = require('electron').remote.require('./bridge');

class EncryptionConfigScreenComponent extends React.Component {
	constructor() {
		super();
		shared.constructor(this);
	}

	componentDidMount() {
		this.isMounted_ = true;
	}

	componentWillUnmount() {
		this.isMounted_ = false;
	}

	initState(props) {
		return shared.initState(this, props);
	}

	async refreshStats() {
		return shared.refreshStats(this);
	}

	componentWillMount() {
		this.initState(this.props);
	}

	componentWillReceiveProps(nextProps) {
		this.initState(nextProps);
	}

	async checkPasswords() {
		return shared.checkPasswords(this);
	}

	renderMasterKey(mk) {
		const theme = themeStyle(this.props.theme);

		const passwordStyle = {
			color: theme.color,
			backgroundColor: theme.backgroundColor,
			border: '1px solid',
			borderColor: theme.dividerColor
		};

		const onSaveClick = () => {
			return shared.onSavePasswordClick(this, mk);
		};

		const onPasswordChange = event => {
			return shared.onPasswordChange(this, mk, event.target.value);
		};

		const password = this.state.passwords[mk.id] ? this.state.passwords[mk.id] : '';
		const active = this.props.activeMasterKeyId === mk.id ? '✔' : '';
		const passwordOk = this.state.passwordChecks[mk.id] === true ? '✔' : '❌';

		return React.createElement(
			'tr',
			{ key: mk.id },
			React.createElement(
				'td',
				{ style: theme.textStyle },
				active
			),
			React.createElement(
				'td',
				{ style: theme.textStyle },
				mk.id
			),
			React.createElement(
				'td',
				{ style: theme.textStyle },
				mk.source_application
			),
			React.createElement(
				'td',
				{ style: theme.textStyle },
				time.formatMsToLocal(mk.created_time)
			),
			React.createElement(
				'td',
				{ style: theme.textStyle },
				time.formatMsToLocal(mk.updated_time)
			),
			React.createElement(
				'td',
				{ style: theme.textStyle },
				React.createElement('input', { type: 'password', style: passwordStyle, value: password, onChange: event => onPasswordChange(event) }),
				' ',
				React.createElement(
					'button',
					{ style: theme.buttonStyle, onClick: () => onSaveClick() },
					_('Save')
				)
			),
			React.createElement(
				'td',
				{ style: theme.textStyle },
				passwordOk
			)
		);
	}

	render() {
		const style = this.props.style;
		const theme = themeStyle(this.props.theme);
		const masterKeys = this.state.masterKeys;
		const containerPadding = 10;

		const headerStyle = Object.assign({}, theme.headerStyle, { width: style.width });

		const containerStyle = Object.assign({}, theme.containerStyle, {
			padding: containerPadding,
			overflow: 'auto',
			height: style.height - theme.headerHeight - containerPadding * 2
		});

		const mkComps = [];
		let nonExistingMasterKeyIds = this.props.notLoadedMasterKeys.slice();

		for (let i = 0; i < masterKeys.length; i++) {
			const mk = masterKeys[i];
			mkComps.push(this.renderMasterKey(mk));

			const idx = nonExistingMasterKeyIds.indexOf(mk.id);
			if (idx >= 0) nonExistingMasterKeyIds.splice(idx, 1);
		}

		const onToggleButtonClick = async () => {
			const isEnabled = Setting.value('encryption.enabled');

			let answer = null;
			if (isEnabled) {
				answer = await dialogs.confirm(_('Disabling encryption means *all* your notes and attachments are going to be re-synchronised and sent unencrypted to the sync target. Do you wish to continue?'));
			} else {
				answer = await dialogs.prompt(_('Enabling encryption means *all* your notes and attachments are going to be re-synchronised and sent encrypted to the sync target. Do not lose the password as, for security purposes, this will be the *only* way to decrypt the data! To enable encryption, please enter your password below.'), '', '', { type: 'password' });
			}

			if (!answer) return;

			try {
				if (isEnabled) {
					await EncryptionService.instance().disableEncryption();
				} else {
					await EncryptionService.instance().generateMasterKeyAndEnableEncryption(answer);
				}
			} catch (error) {
				await dialogs.alert(error.message);
			}
		};

		const decryptedItemsInfo = React.createElement(
			'p',
			{ style: theme.textStyle },
			shared.decryptedStatText(this)
		);
		const toggleButton = React.createElement(
			'button',
			{
				style: theme.buttonStyle,
				onClick: () => {
					onToggleButtonClick();
				}
			},
			this.props.encryptionEnabled ? _('Disable encryption') : _('Enable encryption')
		);

		let masterKeySection = null;

		if (mkComps.length) {
			masterKeySection = React.createElement(
				'div',
				null,
				React.createElement(
					'h1',
					{ style: theme.h1Style },
					_('Master Keys')
				),
				React.createElement(
					'table',
					null,
					React.createElement(
						'tbody',
						null,
						React.createElement(
							'tr',
							null,
							React.createElement(
								'th',
								{ style: theme.textStyle },
								_('Active')
							),
							React.createElement(
								'th',
								{ style: theme.textStyle },
								_('ID')
							),
							React.createElement(
								'th',
								{ style: theme.textStyle },
								_('Source')
							),
							React.createElement(
								'th',
								{ style: theme.textStyle },
								_('Created')
							),
							React.createElement(
								'th',
								{ style: theme.textStyle },
								_('Updated')
							),
							React.createElement(
								'th',
								{ style: theme.textStyle },
								_('Password')
							),
							React.createElement(
								'th',
								{ style: theme.textStyle },
								_('Password OK')
							)
						),
						mkComps
					)
				),
				React.createElement(
					'p',
					{ style: theme.textStyle },
					_('Note: Only one master key is going to be used for encryption (the one marked as "active"). Any of the keys might be used for decryption, depending on how the notes or notebooks were originally encrypted.')
				)
			);
		}

		let nonExistingMasterKeySection = null;

		if (nonExistingMasterKeyIds.length) {
			const rows = [];
			for (let i = 0; i < nonExistingMasterKeyIds.length; i++) {
				const id = nonExistingMasterKeyIds[i];
				rows.push(React.createElement(
					'tr',
					{ key: id },
					React.createElement(
						'td',
						{ style: theme.textStyle },
						id
					)
				));
			}

			nonExistingMasterKeySection = React.createElement(
				'div',
				null,
				React.createElement(
					'h1',
					{ style: theme.h1Style },
					_('Missing Master Keys')
				),
				React.createElement(
					'p',
					{ style: theme.textStyle },
					_('The master keys with these IDs are used to encrypt some of your items, however the application does not currently have access to them. It is likely they will eventually be downloaded via synchronisation.')
				),
				React.createElement(
					'table',
					null,
					React.createElement(
						'tbody',
						null,
						React.createElement(
							'tr',
							null,
							React.createElement(
								'th',
								{ style: theme.textStyle },
								_('ID')
							)
						),
						rows
					)
				)
			);
		}

		return React.createElement(
			'div',
			null,
			React.createElement(Header, { style: headerStyle }),
			React.createElement(
				'div',
				{ style: containerStyle },
				React.createElement(
					'div',
					{ style: { backgroundColor: theme.warningBackgroundColor, paddingLeft: 10, paddingRight: 10, paddingTop: 2, paddingBottom: 2 } },
					React.createElement(
						'p',
						{ style: theme.textStyle },
						React.createElement(
							'span',
							null,
							_('For more information about End-To-End Encryption (E2EE) and advice on how to enable it please check the documentation:')
						),
						' ',
						React.createElement(
							'a',
							{
								onClick: () => {
									bridge().openExternal('https://joplinapp.org/e2ee/');
								},
								href: '#'
							},
							'https://joplinapp.org/e2ee/'
						)
					)
				),
				React.createElement(
					'h1',
					{ style: theme.h1Style },
					_('Status')
				),
				React.createElement(
					'p',
					{ style: theme.textStyle },
					_('Encryption is:'),
					' ',
					React.createElement(
						'strong',
						null,
						this.props.encryptionEnabled ? _('Enabled') : _('Disabled')
					)
				),
				decryptedItemsInfo,
				toggleButton,
				masterKeySection,
				nonExistingMasterKeySection
			)
		);
	}
}

const mapStateToProps = state => {
	return {
		theme: state.settings.theme,
		masterKeys: state.masterKeys,
		passwords: state.settings['encryption.passwordCache'],
		encryptionEnabled: state.settings['encryption.enabled'],
		activeMasterKeyId: state.settings['encryption.activeMasterKeyId'],
		notLoadedMasterKeys: state.notLoadedMasterKeys
	};
};

const EncryptionConfigScreen = connect(mapStateToProps)(EncryptionConfigScreenComponent);

module.exports = { EncryptionConfigScreen };
